//-----------------------------------------------------------------------
// Copyright (c) 2020 Snap Inc.
//-----------------------------------------------------------------------

#include <required.glsl>
#include <std_texture.glsl>

uniform sampler2D baseTex;
uniform vec4      baseTexSize;

uniform float mipLevel;
uniform vec3 lightVec[SAMPLE_COUNT];
uniform mat4 fMipLevelsMat[SAMPLE_COUNT/16];

varying vec2 varTex1;

#include <envProvider/includes/studioUtils.glsl>
#include <envProvider/includes/envProviderUtils.glsl>

#ifdef VERTEX_SHADER

attribute vec2 texture0;

void main() {    
    varTex1 = texture0;

    gl_Position = vec4((texture0 * 2.0 - 1.0), 0.0, 1.0);
}

#endif // VERTEX_SHADER


#ifdef FRAGMENT_SHADER

float getMat4ValByIndex(mat4 mat, int index) {
    int row = index / 4;
    int col = index - row * 4;

    return mat[row][col];
}

float getMipLevel(int index) {
    int indexMat = index / 16;
    int indexVal = index - indexMat * 16;

    return getMat4ValByIndex(fMipLevelsMat[indexMat], indexVal);
}

vec4 filterEnvmap1(vec3 vNormal, mat3 rotationMatrix) {
    float fTotalWeight = 0.0;
    vec3 vPrefilteredColor = vec3(0.0);

    for (int i = 0; i < SAMPLE_COUNT; i++)
    {
        vec3 vLight = lightVec[i];
        float NoL = saturate(dot(vNormal, vLight));

        vLight = rotationMatrix * vLight;

        vec2 uv = calcPanoramicTexCoordsFromDir(vLight);

        float fMipLevel = getMipLevel(i);
        uv = calcSeamlessPanoramicUvsForConvolution(uv, baseTexSize.xy, fMipLevel);
        vPrefilteredColor += decodeRGBD(sc_texture2DLod(baseTex, baseTexSize.xy, uv, fMipLevel)) * NoL;

        fTotalWeight += NoL;
    }

    return encodeRGBD(vPrefilteredColor / fTotalWeight);
}

void main() {
    float MIPToRoughness[6];
    MIPToRoughness[0] = 0.0;
    MIPToRoughness[1] = 0.09;
    MIPToRoughness[2] = 0.25;
    MIPToRoughness[3] = 0.46;
    MIPToRoughness[4] = 0.72;
    MIPToRoughness[5] = 1.0;

    vec3 defaultN = vec3(0.0, 1.0, 0.0);
    vec3 targetN = normalize(calcDirFromPanoramicTexCoords(stretchCoordToOne(varTex1, baseTexSize.zw, mipLevel)));

    // calculate rotation matrix defaultN -> targetN
    mat3 rotationMatrix = getRotationMatrix1(targetN, defaultN);

    gl_FragColor = filterEnvmap1(defaultN, rotationMatrix);
}

#endif // FRAGMENT_SHADER
