//-----------------------------------------------------------------------
// Copyright (c) 2020 Snap Inc.
//-----------------------------------------------------------------------

#ifndef ENV_FROM_CAMERA_UTILS
#define ENV_FROM_CAMERA_UTILS

float ensureIsNotZero(float num) {
    const float EPS = 1e-3;
    float add0 = 1.0 - max(abs(sign(num)), (1.0-EPS));
    return num + add0;
}

vec3 ensureIsNotZero(vec3 num) {
    return vec3(ensureIsNotZero(num.x), ensureIsNotZero(num.y), ensureIsNotZero(num.z));
}

float luma(vec3 c) {
    return dot(c, vec3(0.2126, 0.7152, 0.0722));
}

vec2 putCoordsOnEdge(vec2 uv) {
    vec2 a = uv - 0.5;
    return (vec2(0.5) + sign(a) * 0.5);
}

vec2 stretchCoordToOne(vec2 uv, vec2 origTexturePxSize, float lod) {
    vec2 a = uv - 0.5;
    vec2 b = origTexturePxSize * exp2(lod);
    uv = 0.5 + a + sign(a) * b * 0.5;
    return uv;
}

mat3 getRotationMatrix1(vec3 a, vec3 b) {
    vec3 axis = cross( a, b );

    float cosA = dot( a, b );
    float k = 1.0 / (1.0 + cosA);

    mat3 m3 = mat3( (axis.x * axis.x * k) + cosA,
                    (axis.y * axis.x * k) - axis.z,
                    (axis.z * axis.x * k) + axis.y,
                    (axis.x * axis.y * k) + axis.z,
                    (axis.y * axis.y * k) + cosA,
                    (axis.z * axis.y * k) - axis.x,
                    (axis.x * axis.z * k) - axis.y,
                    (axis.y * axis.z * k) + axis.x,
                    (axis.z * axis.z * k) + cosA
                 );

    return m3;
}

#endif // ENV_FROM_CAMERA_UTILS
